
import java.util.*;
import java.io.PrintWriter;

/** This class performs automatic paragraph numbering
 *
 * By default, the number is government style: 1.2.3.4.5.
 * This can be changed by simply overriding the recalcContextString(),
 * calcNumber(), and isParagraphNumber() methods.
 */
class ParagraphNumberer {

    public class TOCEntry {
        String number;
        String text;
        String status;
        List userAnchors;
    }

    public boolean includeAnchors = true;
    public boolean initialAlpha = false;
    public boolean suppressNumbering = false;

    Vector tableOfContents = new Vector();

    Stack context = new Stack();
    String contextString;
    String userStatus;
    int finalParagraphNumber = 0;
    boolean headingSkipped;

    String lastNumberGiven = null;
    List userAnchors = new ArrayList();
    StringBuffer headingText = new StringBuffer();

    /** Get the next paragraph number
     * @param depth The nesting depth of the paragraph; must be greater than 0.
     *     h1 would have depth 1, h2 would have depth 2, and so on.
     * @return The next paragraph number; it will contain <code>depth</code>
     *     separate numbers, separated by periods.
     */
    public String getNextNumber(int depth) {
        ensureDepth(depth-1);
        recalcContextString();

        finalParagraphNumber++;
        String result = calcNumber();

        if (headingSkipped)
            System.err.println("Warning: heading numbers skipped before "+
                               "paragraph " + result);

        headingSkipped = false;
        headingText.setLength(0);
        userAnchors = new ArrayList();
        lastNumberGiven = result;

        String anchor = "";
        if (includeAnchors)
            anchor = buildAnchor(result);
        return anchor + (suppressNumbering ? "" : result);
    }

    public void appendHeadingText(String text) {
        text = text.replace('\n', ' ');
        text = text.replace('\r', ' ');
        headingText.append(text);
    }

    public void addUserAnchor(String text) {
        userAnchors.add(text);
    }

    public void setUserStatus(String status) {
        this.userStatus = status;
    }

    public void headingTextDone() {
        if (lastNumberGiven == null) return;

        // delete repeated space characters
        String text = headingText.toString();
        text = findAndReplace(text, "&nbsp;", " ");
        text = findAndReplace(text, "  ", " ");
        text = text.trim();

        // add an entry to the table of contents.
        TOCEntry entry = new TOCEntry();
        entry.number = lastNumberGiven;
        entry.text   = text;
        entry.status = userStatus;
        entry.userAnchors = userAnchors;
        tableOfContents.add(entry);

        // reset the variables for the next entry.
        lastNumberGiven = null;
        headingText.setLength(0);
    }


    private void ensureDepth(int contextDepth) {
        while (context.size() < contextDepth) pushDepth();
        while (context.size() > contextDepth) popDepth();
    }

    private void pushDepth() {
        if (finalParagraphNumber == 0) {
            finalParagraphNumber = 1;
            headingSkipped = true;
        }
        context.push(new Integer(finalParagraphNumber));

        finalParagraphNumber = 0;
    }

    private void popDepth() {
        Integer i = (Integer) context.pop();
        finalParagraphNumber = i.intValue();
    }

    protected String calcNumber() {
        int depth = context.size() + 1;
        String finalNumber = getNumber(finalParagraphNumber, depth);
        return contextString + finalNumber + ".";
    }

    protected void recalcContextString() {
        StringBuffer result = new StringBuffer();
        Iterator i = context.iterator();
        Integer num;
        int depth = 0;
        while (i.hasNext()) {
            num = (Integer) i.next();
            result.append(getNumber(num.intValue(), ++depth)).append(".");
        }

        contextString = result.toString();
    }

    protected String getNumber(int num, int depth) {
        if (initialAlpha && depth == 1)
            return (new Character((char) ('A' + num - 1))).toString();
        else
            return Integer.toString(num);
    }

    public boolean isParagraphNumber(String text) {
        for (int i = text.length();   i-- > 0; )
            if (".0123456789".indexOf(text.charAt(i)) == -1)
                return false;
        return true;
    }

    private String makeParaAnchor(String paraNum) {
        return "para_" + paraNum.replace('.', '_');
    }

    public String buildAnchor(String paraNum) {
        paraNum = makeParaAnchor(paraNum);
        return "<a name=\"" + paraNum + "\"></a>";
    }

    private int countDepth(String number) {
        int result = 0;
        for (int i = number.length();  i-- > 0; )
            if (number.charAt(i) == '.')
                result++;
        return result;
    }

    public void printTOC(PrintWriter out) {
        boolean showStatuses = (IncludeCrawler.DEFAULT_DOC_STATUS !=
                                IncludeCrawler.STATUSES_DISABLED);

        out.println("<table cellpadding=2 cellspacing=0 border=0>");
        Iterator i = tableOfContents.iterator();
        while (i.hasNext()) {

            TOCEntry e = (TOCEntry) i.next();

            if (showStatuses) {
                String status = e.status;
                if (status == null) status = IncludeCrawler.FINAL_STATUS;
                status = status.toUpperCase();
                out.print("<tr class=\"" + status + "_TOC_STATUS\">");
                out.print("<td class=\"" + status + "_STATUS\">&nbsp;</td>");
            } else {
                out.print("<tr>");
            }

            int depth = countDepth(e.number);
            int d = depth;
            while (--depth > 0)
                out.print("<td></td>");
            out.print("<td><b>");
            String hyperlink =
                "<a href=\"#" + makeParaAnchor(e.number) + "\">";
            if (e.text == null) {
                out.print(hyperlink);
                out.print(e.number);
                out.println("</a></b></td></tr>");
            } else {
                if (suppressNumbering)
                    out.print("&nbsp;&nbsp;&nbsp;");
                else
                    out.print(e.number);
                out.print("</b>&nbsp;</td><td colspan=");
                out.print(10-d);
                out.print(">");
                out.print(hyperlink);
                out.print(e.text);
                out.println("</a></td></tr>");
            }
        }
        // add a final row which ensures the browser allocates as much
        // space as possible to the final column.
        out.print("<tr>");
        if (showStatuses) out.print("<td>&nbsp;</td>");
        for (int c = MAX_COLUMNS;   c-- > 1; )
            out.print("<td>&nbsp;</td>");
        out.print("<td>");
        for (int c = 80;  c-- > 0; ) out.print("&nbsp; ");
        out.println("&nbsp;</td></tr>");

        out.println("</table>");
    }
    private static final String TAB = "<td></td>";
    private static final int MAX_COLUMNS = 10;


    public TOCEntry getSectionForRef(String ref) {
        Iterator i = tableOfContents.iterator();
        TOCEntry e;
        while (i.hasNext()) {
            e = (TOCEntry) i.next();
            if (e.userAnchors != null && e.userAnchors.contains(ref))
                return e;
        }
        return null;
    }

    public String getNumberForRef(String ref, boolean hyperlink) {
        TOCEntry section = getSectionForRef(ref);
        if (section == null) return null;
        String result =
            (suppressNumbering ? "&quot;" + section.text + "&quot;"
                               : section.number);
        if (result.endsWith("."))
            result = result.substring(0, result.length()-1);

        if (hyperlink)
            result = "<a href=\"#" + makeParaAnchor(section.number) + "\">"
                + result + "</a>";

        return result;
    }

    public String getTextForRef(String ref) {
        TOCEntry section = getSectionForRef(ref);
        return (section == null ? null : section.text);
    }

    public void startInitialAlpha() {
        context.clear();
        finalParagraphNumber = 0;
        initialAlpha = true;
    }


    /**
       A simple find and replace.
       @param text The original string.
       @param find The text to look for.
       @param replace What to replace the found text with.
       @return The new string.
    */
    public static final String findAndReplace(String text,
                                              String find,
                                              String replace)
    {
        if ((text == null) ||
            (find == null) ||
            (replace == null)) {
                throw new NullPointerException
                    ("findAndReplace doesn't work on nulls.");
            }

        // handle degenerate case: if no replacements need to be made,
        // return the original text unchanged.
        int replaceStart = text.indexOf(find);
        if (replaceStart == -1) return text;

        int findLength = find.length();
        StringBuffer toReturn = new StringBuffer();

        while (replaceStart != -1) {
            toReturn.append(text.substring(0, replaceStart)).append(replace);
            text = text.substring(replaceStart+findLength);
            replaceStart = text.indexOf(find);
        }

        toReturn.append(text);
        return toReturn.toString();

    }

}
